from archytas.tool_utils import tool


@tool()
def fib_n(n: int) -> int:
    """
    generate the nth fibonacci number

    Args:
        n (int): The index of the fibonacci number to get

    Returns:
        int: The nth fibonacci number

    Examples:
        >>> fib_n(10)
        55
        >>> fib_n(20)
        6765
    """
    n0 = 0
    n1 = 1
    for _ in range(n):
        n0, n1 = n1, n0 + n1

    return n0


@tool()
def example_tool(arg1: int, arg2: str = "", arg3: dict = None) -> int:
    """
    Simple 1 sentence description of the tool

    More detailed description of the tool. This can be multiple lines.
    Explain more what the tool does, and what it is used for.

    Args:
        arg1 (int): Description of the first argument.
        arg2 (str): Description of the second argument. Defaults to ''.
        arg3 (dict): Description of the third argument. Defaults to {}.

    Returns:
        int: Description of the return value

    Examples:
        >>> example_tool(1, 'hello', {'a': 1, 'b': 2})
        3
        >>> example_tool(2, 'world', {'a': 1, 'b': 2})
        4
    """
    return 42


@tool()
def calculator(expression: str) -> float:
    """
    A simple calculator tool. Can perform basic arithmetic

    Expressions must contain exactly:
    - one left operand. Can be a float or integer
    - one operation. Can be one of + - * / ^ %
    - one right operand. Can be a float or integer

    multiple chained operations are not currently supported.

    Expressions may not contain parentheses, or multiple operations.
    If you want to do a complex calculation, you must do it in multiple steps.

    Args:
        expression (str): A string containing a mathematical expression.

    Returns:
        float: The result of the calculation

    Examples:
        >>> calculator('22/7')
        3.142857142857143
        >>> calculator('3.24^2')
        10.4976
        >>> calculator('3.24+2.5')
        5.74
    """

    # ensure that only one operation is present
    ops = [c for c in expression if c in "+-*/^%"]
    if len(ops) > 1:
        raise ValueError(
            f"Invalid expression, too many operators. Expected exactly one of '+ - * / ^ %', found {', '.join(ops)}"
        )
    if len(ops) == 0:
        raise ValueError(
            f"Invalid expression, no operation found. Expected one of '+ - * / ^ %'"
        )

    op = ops[0]

    _a, _b = expression.split(op)
    a = float(_a)
    b = float(_b)

    if op == "+":
        return a + b
    elif op == "-":
        return a - b
    elif op == "*":
        return a * b
    elif op == "/":
        return a / b
    elif op == "^":
        return a**b
    elif op == "%":
        return a % b


# TODO: still not great since have to include cls in the signature
class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


import math


# @tool()
class Math:
    def __init__(self):
        pass

    # @tool
    @staticmethod
    def sin(x: float) -> float:
        """
        Calculate the sine of x

        Args:
            x (float): The angle in radians

        Returns:
            float: The sine of x
        """
        return math.sin(x)

    # @tool
    @staticmethod
    def cos(x: float) -> float:
        """
        Calculate the cosine of x

        Args:
            x (float): The angle in radians

        Returns:
            float: The cosine of x
        """
        return math.cos(x)

    # @tool
    @staticmethod
    def tan(x: float) -> float:
        """
        Calculate the tangent of x

        Args:
            x (float): The angle in radians

        Returns:
            float: The tangent of x
        """
        return math.tan(x)

    # @tool
    @classproperty  # TODO: still not great since have to include cls in the signature
    def pi(*_) -> float:
        """
        Get the value of pi

        Returns:
            float: The value of pi
        """
        return math.pi


from random import random


class Jackpot:
    """
    A simple slot machine game

    Start with 100 chips, and make bets to try and win more chips.
    """

    def __init__(
        self,
        chips: float = 100,
        win_table: list[tuple[float, float]] = [
            (0.01, 20),
            (0.02, 10),
            (0.05, 4.5),
            (0.2, 1.25),
        ],
    ):
        self._initial_chips = chips
        self.chips = chips
        self.win_table = win_table

    @tool()
    def spin(self, bet: float) -> float:
        """
        Spin the slot machine

        Args:
            bet (float): The amount to bet. Must be less than or equal to the current amount of chips in your wallet

        Returns:
            float: The amount won or lost
        """
        if bet > self.chips:
            raise ValueError(
                f"Bet must be less than or equal to the current winnings. Bet: {bet}, Winnings: {self.winnings}"
            )

        spin = random()
        total_prob = 0
        multiplier = -1
        for prob, win_multiplier in self.win_table:
            total_prob += prob
            if spin <= total_prob:
                multiplier = win_multiplier
                break

        self.chips += bet * multiplier
        return bet * multiplier

    @tool()
    def get_chips(self) -> float:
        """
        Get the current amount of chips in your wallet

        Returns:
            float: The current amount of chips in your wallet
        """
        return self.chips

    @tool()
    def reset(self):
        """
        Reset the game back to the initial number of chips
        """
        self.chips = self._initial_chips


class ModelSimulation:
    """
    Simple example of a SIR model simulation
    """

    def __init__(self, dt=0.1):
        self._default_parameters = {
            "beta": 0.002,
            "gamma": 0.1,
            "S": 990,
            "I": 10,
            "R": 0,
        }
        self.parameters = self._default_parameters.copy()
        self.dt = dt

    @tool()
    def get_model_parameters(self) -> dict:
        """
        Get the model parameters

        Returns:
            dict: The model parameters in the form {param0: value0, param1: value1, ...}

        """
        return self.parameters

    @tool()
    def set_model_parameters(self, update: dict):
        """
        Set some or all of the model parameters

        Args:
            update (dict): The parameters to update. Should be a dict of the form {param0: value0, param1: value1, ...}. Only the parameters specified will be updated.
        """
        self.parameters.update(update)

    @tool()
    def run_model(self, steps: int = 100) -> dict:
        """
        Run the model for a number of steps

        Args:
            steps (int): The number of steps to run the model for. Defaults to 100.

        Returns:
            dict: The model results in the form {param0: value0, param1: value1, ...}
        """
        S_new, I_new, R_new = (
            self.parameters["S"],
            self.parameters["I"],
            self.parameters["R"],
        )
        beta, gamma = self.parameters["beta"], self.parameters["gamma"]
        population = S_new + I_new + R_new

        for _ in range(steps):
            S_old, I_old, R_old = S_new, I_new, R_new
            dS = -beta * S_old * I_old
            dI = beta * S_old * I_old - gamma * I_old
            dR = gamma * I_old

            S_new = max(0, min(S_old + self.dt * dS, population))
            I_new = max(0, min(I_old + self.dt * dI, population))
            R_new = max(0, min(R_old + self.dt * dR, population))

            # Ensure the total population remains constant
            total_error = population - (S_new + I_new + R_new)
            R_new += total_error

        self.parameters["S"], self.parameters["I"], self.parameters["R"] = (
            S_new,
            I_new,
            R_new,
        )
        return self.parameters

    @tool()
    def reset_model(self):
        """
        Reset the model to the initial parameters
        """
        self.parameters = self._default_parameters.copy()


@tool()
def ObservablePlot(code: str):
    """
    Create an observable plot in code.

    Args:
        code (str): The Observable Plot code

    """


class PlannerTool:
    """
    A tool for helping to make long term plans. After a plan is made, the system will remind you of the plan as you execute each of the steps.
    """

    def __init__(self):
        self.current_plan = None
        self.progress = None

    @tool()
    def make_new_plan(self, draft: list[str]) -> None:
        """
        Start the plan making process.

        This starts a conversation with the system to refine and revise a plan until it is robust and well thought out.

        Args:
            draft (list[str]): An initial list of steps to include in the plan. These can be revised later.
        """
        self.current_plan = draft
        import pdb

        pdb.set_trace()
        ...

    @tool()
    def revise_single_step(self, i: int, replacement: str) -> None:
        """
        Revise a single step in the current plan

        Args:
            i (int): The index of the step to revise
            replacement (str): The new step to replace the old step with
        """
        raise NotImplementedError("TODO")

    # @tool()
    # def revise_plan(self, updates:dict[int,str]) -> None: ...

    # @tool()
    # def finalize_plan():...


from .tool_utils import AgentRef


@tool()
def pirate_subquery(query: str, agent: AgentRef) -> str:
    """
    Runs a subquery using a oneshot agent in which answers will be worded like a pirate.

    Args:
        query (str): The query to run against the agent.

    Returns:
        str: Result of the subquery in pirate vernacular.

    """
    prompt = """
    You are an pirate. Answer all questions truthfully using pirate vernacular.
    """
    return agent.oneshot(prompt=prompt, query=query)



import asyncio
import inspect
import json
import sys
import logging
import traceback
import typing
import uuid
from langchain_core.messages import AIMessage, HumanMessage, ToolMessage, ToolCall

from .agent import Agent, BaseMessage, SystemMessage, AgentResponse
from .prompt import build_prompt, build_all_tool_names
from .tools import ask_user
from .tool_utils import make_tool_dict, sanitize_toolname
from .models.base import BaseArchytasModel
from .utils import extract_json


logger = logging.Logger("archytas")


class Undefined:
    ...


class FailedTaskError(Exception):
    def __init__(
            self,
            message: str,
            last_error: typing.Optional[BaseException] = None,
            tool_call_id: typing.Optional[str] = None,
            extra: typing.Optional[str] = None,
            *extra_args: object
        ) -> None:
        self.message = message
        self.last_error = last_error
        self.tool_call_id = tool_call_id
        self.extra = extra
        super().__init__(*extra_args)

    def __str__(self) -> str:
        err_str = f"""
Error message:
  {self.message}
"""
        if self.last_error:
            err_str += f"""
During processing, the following error was raised:
  {self.last_error}
"""
        if self.extra:
            err_str += f"""
Extra details:
  {self.extra}
"""
        return err_str


def catch_failure(fn):
    def handle_error(messages: list[BaseMessage], error: Exception):
        last_ai_message = None
        seen_tool_message_ids = set()
        for message in messages[::-1]:
            match message:
                case AIMessage():
                    last_ai_message = message
                    break
                case ToolMessage():
                    seen_tool_message_ids.add(message.id)
        if not last_ai_message:
            return
        if last_ai_message.tool_calls:
            missing_tool_ids = set(tool_call["id"] for tool_call in last_ai_message.tool_calls if tool_call["id"] not in seen_tool_message_ids)
            for missing_tool_id in missing_tool_ids:
                error_message = ToolMessage(
                    content=str(error),
                    tool_call_id=missing_tool_id,
                )
                messages.append(error_message)

    async def inner_async(self, *args, **kwargs):
        try:
            return await fn(self, *args, **kwargs)
        except FailedTaskError as failed_task:
            handle_error(self.messages, failed_task)
            raise
        except Exception as error:
        # TODO: Is it good to handle other errors here too? Probably, but flow can be tricky.
        # TODO: Would need to find dangling tools and generate tool messages for any dangling tool calls
            handle_error(self.messages, error)
            raise

    def inner(self, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except FailedTaskError as failed_task:
            handle_error(self.messages, failed_task)
            raise

    if inspect.iscoroutine(fn) or inspect.iscoroutinefunction(fn) or inspect.isawaitable(fn):
        return inner_async
    else:
        return inner


class LoopController:
    PROCEED = 0
    STOP_SUCCESS = 1
    STOP_FATAL = 2

    state: int

    def __init__(self) -> None:
        self.reset()

    def set_state(self, new_value):
        self.state = new_value

    def reset(self):
        self.state = 0


class AutoSummarizedToolMessage(ToolMessage):
    """A message that replaces its full tool output with a summary after the ReAct loop is complete."""

    summary: str = ""
    summarized: bool = False

    async def update_content(self):
        if not self.summarized:
            self.content=self.summary
            self.summarized = True

class ReActAgent(Agent):
    def __init__(
        self,
        *,
        model: BaseArchytasModel | None = None,
        api_key: str | None = None,
        tools: list = None,
        allow_ask_user: bool = True,
        max_errors: int | None = 3,
        max_react_steps: int | None = None,
        thought_handler: typing.Callable | None = Undefined,
        messages: typing.Optional[list[BaseMessage]] | None = None,
        **kwargs,
    ):
        """
        Create a ReAct agent

        Args:
            model (BaseArchytasModel): The model to use. Defaults to OpenAIModel(model_name="gpt-4o").
            api_key (str, optional): The LLM provider API key to use. Defaults to None. If None, the provider will use the default environment variable (e.g. OPENAI_API_KEY).
            tools (list): A list of tools to use. Defaults to None. If None, only the system tools (final_answer, fail_task) will be used.
            allow_ask_user (bool): Whether to include the ask_user tool, which allows the model to ask the user for clarification. Defaults to True.
            max_errors (int, optional): The maximum number of errors to allow during a task. Defaults to 3.
            max_react_steps (int, optional): The maximum number of steps to allow during a task. Defaults to infinity.
            thought_handler (function, optional): Hook to control logging/output of the thoughts made in the middle of a react loop. Set to None to disable, or leave default of Undefined to
                    print to terminal. Otherwise expects a callable function with the signature of `func(thought: str, tool_name: str, tool_input: str) -> None`.
            messages (list[BaseMessage], optional): A list of messages to start the agent with. Defaults to None.
        """
        # create a dictionary for looking up tools by name
        tools = tools or []
        if allow_ask_user:
            tools.append(ask_user)
        tools.append(self)
        self._raw_tools = tools
        self.tools = make_tool_dict(tools)
        self.current_query = None

        if thought_handler is Undefined:
            self.thought_handler = self.thought_callback
        else:
            self.thought_handler = thought_handler

        # check that the tools dict keys match the list of generated tool names
        names, keys = sorted(build_all_tool_names(tools)), sorted([*self.tools.keys()])
        assert (
            names == keys
        ), f"Internal Error: tools dict keys does not match list of generated tool names. {names} != {keys}"

        # create the prompt with the tools, and initialize the agent
        self.prompt = build_prompt(tools)
        super().__init__(model=model, prompt=self.prompt, api_key=api_key, messages=messages, **kwargs)

        # react settings
        self.max_errors = max_errors or float("inf")
        self.max_react_steps = max_react_steps or float("inf")

        # number of errors and steps during current task
        self.errors = 0
        self.steps = 0

        # keep track of the last tool used (for error messages)
        self.last_tool_name = ""

    def update_prompt(self):
        self.prompt = build_prompt(self._raw_tools)
        if self.model.MODEL_PROMPT_INSTRUCTIONS:
            self.prompt += "\n\n" + self.model.MODEL_PROMPT_INSTRUCTIONS
        self.system_message = SystemMessage(content=self.prompt)

    def disable(self, *tool_names):
        if len(tool_names) == 0:
            return
        tool = None
        for tool_name in tool_names:
            if tool_name in self.tools:
                tool = self.tools[tool_name]
            elif "." in tool_name:
                matches = [name for name in self.tools.keys() if name.endswith(f".{tool_name}")]
                if len(matches) > 1:
                    raise ValueError(f"Ambiguous name: Multiple tools called '{tool_name}'")
                tool = self.tools[matches[0]]
            if tool is None:
                # If tool is not enabled/included, it can't be disabled, so skip.
                continue
            if inspect.ismethod(tool):
                setattr(tool.__func__, '_disabled', True)
            else:
                setattr(tool, '_disabled', True)

        self.update_prompt()
        if self.model:
            self.model.set_tools(self.tools)

    def thought_callback(self, thought: str, tool_name: str, tool_input: str) -> None:
        if self.verbose:
            # TODO: better coloring
            self.print(
                f"thought: {thought}\ntool: {tool_name}\ntool_input: {tool_input}\n"
            )

    def react(self, query: str, react_context:dict=None) -> str:
        """
        Synchronous wrapper around the asynchronous react_async method.
        """
        return asyncio.run(self.react_async(query, react_context))

    @catch_failure
    async def react_async(self, query: str, react_context:dict=None) -> str:
        """
        Asynchronous react loop function.
        Continually calls tools until a satisfactory answer is reached.
        """

        # reset error and steps counter
        self.errors = 0
        self.steps = 0

        action: AgentResponse | None = None
        reaction: AgentResponse | None = None

        # Set the current query for use in tools, auto context, etc
        self.current_query = query
        action = await self.handle_message(HumanMessage(content=query))

        controller = LoopController()

        # ReAct loop
        while True:
            # Start processing last loops reaction as a new action
            if reaction is not None:
                action = reaction

            if not action.tool_calls:
                if action.text:
                    try:
                        # Check to ensure the content isn't an tool call in the wrong place.
                        action_json = json.loads(action.text)
                        if isinstance(action_json, dict) and "id" in action_json and "name" in action_json and "args" in action_json:
                            action.tool_calls.append(action_json)
                    except json.JSONDecodeError:
                        # Presume content to be a final answer
                        action.tool_calls.append({
                            "id": uuid.uuid4().hex,
                            "name": "final_answer",
                            "args": {
                                "response": action.text
                            }
                        })

            if action.tool_calls:
                for tool_call in action.tool_calls:
                    tool_id = tool_call["id"]
                    tool_name = tool_call["name"]
                    tool_args = tool_call["args"]

                    # exit ReAct loop if agent says final_answer or fail_task
                    if tool_name == "final_answer":
                        await self.summarize_messages()
                        self.debug(
                            event_type="react_final_answer",
                            content={
                                "final_answer": tool_args,
                            }
                        )
                        self.current_query = None
                        # Store final answer as response in message,
                        response = tool_args.get("response", None)
                        if not response:
                            # TODO: Handle this case
                            raise ValueError("The LLM provided an empty message for a final_answer. This is not valid.")
                        self.messages.append(ToolMessage(content=str(response), tool_call_id=tool_id))
                        return response
                    if tool_name == "fail_task":
                        self.current_query = None
                        reason = tool_args.get("reason", "No reason identified.")
                        error = tool_args.get("error", None)
                        raise FailedTaskError(
                            message=reason,
                            extra=error,
                            tool_call_id=tool_id
                        )

                    if self.thought_handler:
                        self.thought_handler(action.text, tool_name, tool_args)


                    # run tool
                    try:
                        tool_fn = self.tools[tool_name]
                    except KeyError:
                        self.messages.append(ToolMessage(
                            content=f'Unknown tool "{tool_name}"\nAvailable tools: {", ".join(self.tools.keys())}',
                            tool_call_id=tool_id,
                        ))
                        continue

                    try:
                        tool_context = {
                            "agent": self,
                            "tool_name": tool_name,
                            "raw_tool": tool_fn,
                            "loop_controller": controller,
                            "react_context": react_context,
                        }
                        tool_self_ref = getattr(tool_fn, "__self__", None)
                        self.debug(
                            event_type="react_tool",
                            content={
                                "tool": tool_name,
                                "input": tool_args,
                            }
                        )
                        tool_output = await tool_fn.run(args=tool_args, tool_context=tool_context, self_ref=tool_self_ref)
                        self.debug(
                            event_type="react_tool_output",
                            content={
                                "tool": tool_name,
                                "input": tool_args,
                                "output": tool_output,
                            }
                        )

                        # Auto-summarize if required
                        if getattr(tool_fn, "autosummarize", False):
                            summary = f"Summary of action: Executed command '{tool_name}' with arguments '{tool_args}'"
                            tool_message = AutoSummarizedToolMessage(
                                content=tool_output,
                                summary=summary,
                                tool_call_id=tool_id,
                            )
                        else:
                            tool_message = ToolMessage(
                                content=tool_output,
                                tool_call_id=tool_id,
                            )

                        # Always add tool response before handling state changes to ensure message history is correct.
                        self.messages.append(tool_message)

                        # Have the agent observe the result, and get the next action
                        if self.verbose:
                            self.display_observation(tool_output)

                        # Log cases when tools override controller state
                        if controller.state != LoopController.PROCEED:
                            self.debug(
                                event_type="react_controller_state",
                                content={
                                    "state": controller.state,
                                }
                            )
                        # Check loop controller to see if we need to stop or error
                        if controller.state == LoopController.STOP_SUCCESS:
                            await self.summarize_messages()
                            self.current_query = None
                            return tool_output
                        if controller.state == LoopController.STOP_FATAL:
                            await self.summarize_messages()
                            self.current_query = None
                            raise FailedTaskError(
                                tool_output,
                                tool_call_id=tool_id
                            )
                    except asyncio.CancelledError:
                        self.messages.append(ToolMessage(
                            content='Execution of this tool was interrupted by the user.',
                            tool_call_id=tool_id
                        ))
                    except Exception as e:
                        self.messages.append(ToolMessage(
                            content=f'error running tool "{tool_name}"\n\n:{e}\n{traceback.format_exception(e)}',
                            tool_call_id=tool_id
                        ))

            # Execute to fetch next step in the ReAct loop
            reaction = await self.execute()

    @staticmethod
    def extract_action(action: dict) -> tuple[str, str, str, bool]:
        """Verify that action has the correct keys. Otherwise, raise an error"""
        assert isinstance(
            action, dict
        ), f"Action must be a json dictionary, got {type(action)}"
        assert "thought" in action, "Action json is missing key 'thought'"
        assert "tool" in action, "Action json is missing key 'tool'"
        assert "tool_input" in action, "Action json is missing key 'tool_input'"

        thought = action["thought"]
        tool = action["tool"]
        tool_input = action["tool_input"]
        helpful_thought = action.get("helpful_thought", True)


        return thought, tool, tool_input, helpful_thought

    def execute(self, additional_messages: list[BaseMessage] = []) -> str:
        """
        Execute the model and return the output (see `Agent.execute()`).
        Keeps track of the number of execute calls, and raises an error if there are too many.
        """
        self.steps += 1
        if self.steps > self.max_react_steps:
            last_tool_id = None
            raise FailedTaskError(
                f"Too many steps ({self.steps} > max_react_steps) during task.\nLast action should have been either final_answer or fail_task. Instead got: {self.last_tool_name}",
                tool_call_id=last_tool_id,
            )
        return super().execute(additional_messages, tools=self.tools)

    def error(self, mesg: str, err: BaseException, tool_id: str | None = None) -> str:
        """error handling. If too many errors, break the ReAct loop. Otherwise tell the agent, and continue"""
        self.errors += 1
        if self.errors >= self.max_errors:
            raise FailedTaskError(
                f"Too many errors during task. Last error: {mesg}",
                last_error=err,
                tool_call_id=tool_id
            )
        if self.verbose:
            if self.rich_print:
                self.print(f"[red]error: {mesg}[/red]", file=sys.stderr)
            else:
                self.print(f"error: {mesg}", file=sys.stderr)

    def display_observation(self, observation):
        """
        Display the observation. Can be overridden by subclasses to display the observation in different ways.
        """
        self.print(f"observation: {observation}\n")

    async def summarize_messages(self):
        """Summarizes and self-summarizing tool messages."""
        for message in self.messages:
            if isinstance(message, AutoSummarizedToolMessage) and not message.summarized:
                await message.update_content()



import os
import re
from functools import lru_cache
from typing import Any, Optional, Annotated
from langchain_openai.chat_models import ChatOpenAI
from langchain_core.messages import FunctionMessage, AIMessage
from langchain.tools import StructuredTool

from openai import AuthenticationError as OpenAIAuthenticationError, APIError, APIConnectionError, RateLimitError, OpenAIError
from .base import BaseArchytasModel, ModelConfig, set_env_auth
from ..exceptions import AuthenticationError, ExecutionError

DEFERRED_TOKEN_VALUE = "***deferred***"

class OpenAIModel(BaseArchytasModel):
    tool_descriptions: dict[str, str]

    @property
    def MODEL_PROMPT_INSTRUCTIONS(self):

        tool_desc = ["The following tools are available:"]
        for tool_name, tool_description in self.tool_descriptions.items():
            tool_desc.append(
                f"""\
{tool_name}:
    {tool_description}
"""
            )
        return "\n------\n".join(tool_desc)

    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)
        self.tool_descriptions = {}

    def auth(self, **kwargs) -> None:
        auth_token = None
        if 'api_key' in kwargs:
            auth_token = kwargs['api_key']
        elif 'api_key' in self.config:
            auth_token = self.config['api_key']
        if not auth_token:
            auth_token = DEFERRED_TOKEN_VALUE
        set_env_auth(OPENAI_API_KEY=auth_token)

        # Replace local auth token from value from environment variables to allow fetching preset auth variables in the
        # environment.
        auth_token = os.environ.get('OPENAI_API_KEY', DEFERRED_TOKEN_VALUE)

        if auth_token != DEFERRED_TOKEN_VALUE:
            self.config['api_key'] = auth_token
        # Reset the openai client with the new value, if needed.
        if getattr(self, "model", None):
            self.model.openai_api_key._secret_value = auth_token
            self.model.client = None
            self.model.async_client = None

            # This method reinitializes the clients
            self.model.validate_environment()

    def convert_tools(self, archytas_tools: tuple[tuple[str, Any], ...])-> "list[StructuredTool]":
        tools = super().convert_tools(archytas_tools)
        self.tool_descriptions = {}
        for tool in tools:
            if len(tool.description) > 1024:
                self.tool_descriptions[tool.name] = tool.description
                tool.description = f"The description for this tool, `{tool.name}`, can be found in the system message on this call."
        return tools

    def initialize_model(self, **kwargs):
        try:
            return ChatOpenAI(model=self.config.get("model_name", "gpt-4o"))
        except (APIConnectionError, OpenAIError) as err:
            if not self.config.get('api_key', None):
                raise AuthenticationError("OpenAI API Key not set")
            else:
                raise AuthenticationError("OpenAI Authentication Error") from err

    def ainvoke(self, input, *, config=None, stop=None, **kwargs):
        if self.model is None or not getattr(self.model, 'openai_api_key', None):
            raise AuthenticationError("OpenAI API Key missing")
        if "o3" in self.model.model_name.lower():
            # o3 doesn't accept a temperature keyword on invoke
            kwargs.pop("temperature")
        return super().ainvoke(input, config=config, stop=stop, **kwargs)

    def handle_invoke_error(self, error: BaseException):
        if isinstance(error, OpenAIAuthenticationError):
            raise AuthenticationError("OpenAI Authentication Error") from error
        elif isinstance(error, RateLimitError):
            raise ExecutionError(error.message) from error
        elif isinstance(error, (APIConnectionError, OpenAIError)) and not self.model.openai_api_key:
            raise AuthenticationError("OpenAI Authentication Error") from error
        else:
            raise error



from langchain_ollama import ChatOllama, OllamaLLM
from langchain_core.messages import AIMessage, ToolMessage, HumanMessage
from pydantic import BaseModel as PydanticModel, Field

from .base import BaseArchytasModel, EnvironmentAuth, ModelConfig



class OllamaModel(BaseArchytasModel):

    MODEL_PROMPT_INSTRUCTIONS = """
If you receieve a message with a role of `tool` in response to a tool being called, the user has not seen the content
of that message. Those messages are generated by the system and sent to you for review as part of the ReAct loop. Please
treat such messages as a response from a third party that provide more information for you to use to help achieve the
user's task.
Be sure to always use the `final_answer` tool to report back to the user.
"""

    def auth(self, **kwargs) -> None:
        return

    def initialize_model(self, **kwargs):
        # Dummy tool in required to enable "tool mode" on the model
        return ChatOllama(model=self.config.get("model_name", "llama3"))



import json
from typing import Callable
from archytas.tool_utils import get_tool_prompt_description, get_tool_names
from .message_schemas import ToolUseRequest

tool_use_schema = json.dumps(ToolUseRequest.model_json_schema(), indent=2)

def prelude() -> str:
    return f'''\
You are the ReAct (Reason & Action) assistant. You act as an interface between a user and the system.

Your job is to help the user to complete their tasks by calling tools, evaluating the results and communicating with the user.
The user will communicate with you directly in natural language, however you can only respond to queries by calling tools.
Calling tools requires that you respond with properly formatted JSON objects that fit the schema described below.
As the only action you can take is to call a tool, this means that every response you make must be a JSON object.
Any plain text not part of a JSON object will be ignored or will cause errors.

You will be provided tools that will allow communication with the user. All communication should be done via those tools.

Completing the user's task may require calling several tools in a chain or loop. This is fine. You can call as many tools
as you feel is necessary to complete the task, within reason. If you cannot make progress towards the task, or if calling
the provided tools is not getting you closer to completing the task, you should communcate this to the user by failing the
task. Only ever run one tool at a time and review the output of running the tool before deciding upon the next tool to call.
'''


def tool_intro() -> str:
    return "# Tools\nYou have access to the following tools which can help you in your job:"

def system_tools() -> str:
    return f"""
planning:
    The planning tool is used as a scratchpad for you to record notes or plans related to completing the user's task.
    This will not result in any changes in the system or provide any new information, however the information provided will
    be included during future loops. This can be used to provide a plan so you remain focused, or anything else you need
    to remember for a future point.
    Note: This action is not without cost, and so should only be used when deemed helpful. If the task can be completely
    resolved in one or two iterations, you likely should not use this tool.
    _input_: A string representing any information you want to store that may be helpful for you to complete the task.

final_answer:
    The final_answer tool is used to indicate that you have completed the task. You should use this tool to communicate the final answer to the user.
    _input_: A string representing a human-readable reply that conveys the final answer to the user's request, task or question.

fail_task
    The fail_task tool is used to indicate that you have failed to complete the task. You should use this tool to communicate the reason for the failure to the user. Do not call this tool unless you have given a good effort to complete the task.
    _input_: A plain text explanation of the reason for the failure, along with the raw root cause reason of the error for debugging.
""".strip()


# TODO: there should be some way to give an example relevant to the environment/tools...
#      or use a system tool for the example
def formatting(tool_names: list[str], *, ask_user: bool) -> str:
    tool_names += ["final_answer", "fail_task"]
    quoted_tool_list =", ".join(f"`{tool_name}`" for tool_name in tool_names)
    return f"""
# Formatting
Every response you generate should EXACTLY follow this JSON schema:
```jsonschema
{tool_use_schema}
```
The tools available are: {quoted_tool_list}

In the response text, you should answer using markdown encoding of the json and as such precede the JSON payload with "```json" and follow it with "```" like so:
```json
{{
  "thought": "This is what I'm thinking",
  "tool": "final_answer",
  "tool_input": "This is my final answer to the user",
}}
```

When specifying a tool, be sure to use the entire name, including any preceding or trailing identifiers.

You can communicate with the user via the {'`final_answer` tool, or the `ask_user` tool' if ask_user else '`final_answer` tool'}.
`tool_input` must be a valid JSON value (i.e. null, string, number, boolean, array, or object).
The input type will depend on which tool you select, so make sure to follow the instructions for each tool.

For example, if the user asked you what the square-root of 2, you would use the calculator like so:
```json
{{
    "thought": "I need to use the calculator to find the square-root of 2.",
    "tool": "calculator",
    "tool_input": "2^0.5"
}}
```
""".strip()


def notes(*, ask_user: bool) -> str:
    return f"""
# Notes
- assume any time based knowledge you have is out of date, and should be looked up, if possible. Things like the current date, current world leaders, celebrities ages, etc.
- You are not very good at arithmetic, so you should generally use tools to do arithmetic for you.
- The user may or may not see your thoughts. If you want to communicate something to the user, it should be via the {'ask_user or final_answer tools' if ask_user else 'final_answer tool'}.
""".strip()


def build_prompt(tools: list[Callable]) -> str:
    """
    Build the prompt for the ReAct agent

    Args:
        tools (list): A list of tools to use. Each tool should have the @tool decorator. applied.

    Returns:
        str: The prompt for the ReAct agent
    """
    # collect all the tool names (including class.method names)
    # tool_names = build_all_tool_names(tools)

    # # check if the ask user prompt is in the list of tools
    # ask_user = "ask_user" in tool_names

    # chunks = [prelude(), tool_intro()]
    # for tool in tools:
    #     chunks.append(get_tool_prompt_description(tool))
    # chunks.append(system_tools() + "\n")
    # chunks.append(formatting(tool_names, ask_user=ask_user) + "\n")
    # chunks.append(notes(ask_user=ask_user))
    # return "\n\n".join(chunks)
    # return prelude()
    return f'''\
You are the ReAct (Reason & Action) assistant. You act as an interface between a user and the system.

Your job is to help the user to complete their tasks by calling tools, evaluating the results and communicating with the user.
The user may not see the results of executing the tools, so be sure to communicate the important results/outputs from the tool
executions back to the user following the execution.

Completing the user's task may require calling several tools in a chain or loop. This is fine. You can call as many tools
as you feel is necessary to complete the task, within reason. If you cannot make progress towards the task, or if calling
the provided tools is not getting you closer to completing the task, you should communcate this to the user by failing the
task. Only ever run one tool at a time and review the output of running the tool before deciding upon the next tool to call.

If you are able to provide your thoughts via response text separate from calling a tool. Please do so, explaining your thoughts
as to reasoning behind the call. The user may or may not see these thoughts.
'''


def build_all_tool_names(tools: list[Callable]) -> list[str]:
    """
    Build a list of tool names from a list of tools

    Args:
        tools (list): A list of tools to use. Each tool should have the @tool decorator. applied.

    Returns:
        list: A list of tool names
    """
    tool_names = get_tool_names(tools)
    return tool_names



import json
import re
import logging

from anthropic import AuthenticationError as AnthropicAuthenticError, RateLimitError
from langchain_anthropic.chat_models import ChatAnthropic
from langchain_core.messages import AIMessage, SystemMessage
from pydantic import BaseModel as PydanticModel, Field

from archytas.agent import AIMessage, BaseMessage

from .base import BaseArchytasModel, EnvironmentAuth, ModelConfig
from ..exceptions import AuthenticationError, ExecutionError

class DummyTool(PydanticModel):
    """
    Dummy Tool
    """
    input: str = Field(..., description="input")


class AnthropicModel(BaseArchytasModel):
    DEFAULT_MODEL: str = "claude-3-5-sonnet-latest"
    api_key: str = ""
    tool_name_map: dict
    rev_tool_name_map: dict

    def __init__(self, config: ModelConfig, **kwargs) -> None:
        super().__init__(config, **kwargs)
        self.last_messages: list[BaseMessage] = None
        self.tool_name_map = {}
        self.rev_tool_name_map = {}

    def auth(self, **kwargs) -> None:
        if 'api_key' in kwargs:
            self.api_key = kwargs['api_key']
        elif 'api_key' in self.config:
            self.api_key = self.config['api_key']
        if not self.api_key:
            raise AuthenticationError("No auth credentials found.")

    def initialize_model(self, **kwargs):
        return ChatAnthropic(model=self.config.get("model_name", self.DEFAULT_MODEL), api_key=self.api_key, max_tokens=self.config.get("max_tokens", 4096))

    def _preprocess_messages(self, messages):
        from ..agent import AutoContextMessage, ContextMessage
        output = []

        system_messages = []
        # Combine all system/context/autocontext messages into a single initial system message
        for message in messages:
            match message:
                case SystemMessage() | ContextMessage() | AutoContextMessage():
                    system_messages.append(message.content)
                case _:
                    output.append(message)
        # Condense all context/system messages into a single first message as required by Anthropic
        output.insert(0, SystemMessage(content="\n".join(system_messages)))
        self.last_messages = [msg.model_copy(deep=True) for msg in output]
        return output

    def handle_invoke_error(self, error: BaseException):
        if isinstance(error, AnthropicAuthenticError):
            raise AuthenticationError("Anthropic Authentication Error") from error
        # TODO: Retry with delay on rate limit errors?
        # elif isinstance(error, RateLimitError):
        #     raise
        else:
            if self.last_messages:
                message_output = [msg.model_dump() for msg in self.last_messages]
                logging.warning(
                    "An exception has occurred. Below are the messages that were sent to in the most recent request:\n" +
                    json.dumps(message_output, indent=2)
                )
            raise
